from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.parser import INLINE_BLOCK
from oelint_parser.rpl_regex import RegexRpl

from oelint_adv.cls_rule import Rule


class VarDuplicates(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.duplicate',
                         severity='warning',
                         message='<FOO>')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for c in ['DEPENDS', 'RDEPENDS']:
            items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                      attribute=Variable.ATTR_VAR, attributeValue=c)
            _items = {}
            for i in items:
                for x in [y for y in i.get_items(versioned=True) if y]:
                    machine_mods = i.SubItems
                    machine_mods_cleaned = '_'.join(
                        sorted([x for x in machine_mods if x not in ['append', 'prepend', 'remove']]))
                    m = RegexRpl.match('.*(?P<dyn>/dynamic-layers/.*?/).*', i.Origin)
                    if m:
                        machine_mods_cleaned += f'_{m.group("dyn")}'
                    if machine_mods_cleaned not in _items:
                        _items[machine_mods_cleaned] = []
                    _operations = i.AppendOperation()
                    if x in _items[machine_mods_cleaned]:
                        if x == INLINE_BLOCK:
                            continue
                        if not any(x in ['append', 'prepend', ' += ', ' =+ '] for x in _operations):
                            _items[machine_mods_cleaned] = [x]
                        else:
                            res += self.finding(i.Origin, i.InFileLine,
                                                'Item \'{x}\' was added multiple times to {c}'.format(x=x, c=c))
                    else:
                        if any(x in ['append', 'prepend', ' += ', ' =+ '] for x in _operations):
                            _items[machine_mods_cleaned].append(x)
                        elif 'remove' in _operations:
                            if x in _items[machine_mods_cleaned]:
                                _items[machine_mods_cleaned].remove(x)  # pragma: no cover
                        else:
                            _items[machine_mods_cleaned] = [x]
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class VarSuggestedExists(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.var.suggestedvar',
                         severity='info',
                         message='<FOO>',
                         onappend=False,
                         appendix=CONSTANTS.VariablesSuggested)

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        _is_pkg_group = stash.IsPackageGroup(_file)
        all_items: List[Variable] = stash.GetItemsFor(filename=_file,
                                                      classifier=Variable.CLASSIFIER,
                                                      attribute=Variable.ATTR_VAR,
                                                      attributeValue=CONSTANTS.VariablesSuggested)
        for var_ in CONSTANTS.VariablesSuggested:
            if var_ == 'BBCLASSEXTEND':
                # this is better covered by oelint_adv/rule_base/rule_vars_bbclassextends.py
                continue
            if _is_pkg_group and var_ in ['LICENSE', 'CVE_PRODUCT']:
                continue
            items = [x for x in all_items if x.VarName == var_]
            if not any(items):
                res += self.finding(_file, 0, 'Variable \'{var}\' should be set'.format(var=var_), appendix=var_)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarsVirtualProvDep(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.virtual',
                         severity='error',
                         message='{var} can not contain virtual/ items')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue=['RDEPENDS', 'RPROVIDES'])

        for i in items:
            if any(x.startswith('virtual/') for x in i.get_items() if 'remove' not in i.SubItems):
                res += self.finding(i.Origin, i.InFileLine, override_msg=self.Msg.format(var=i.VarName))
        return res



from typing import List, Tuple

from oelint_parser.cls_item import FlagAssignment, Function
from oelint_parser.cls_stash import Stash
from oelint_parser.constants import CONSTANTS

from oelint_adv.cls_rule import Rule


class TaskDocStrings(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.task.docstrings',
                         severity='info',
                         message='Every custom task should have a doc string set by task[doc] = \'\'')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for item in stash.GetItemsFor(filename=_file, classifier=Function.CLASSIFIER):
            if item.FuncName in CONSTANTS.FunctionsKnown or item.FuncName in ['', 'anonymous']:
                # Skip for builtin tasks or anonymous python functions
                continue
            if item.IsAppend():
                # In case it's an append operation, there has to be an original
                # so don't raise any warnings here
                continue
            _ta: List[FlagAssignment] = stash.GetItemsFor(filename=_file,
                                                          classifier=FlagAssignment.CLASSIFIER,
                                                          attribute=FlagAssignment.ATTR_NAME,
                                                          attributeValue=item.FuncName)
            if not any(x.Flag == 'doc' for x in _ta):
                res += self.finding(item.Origin, item.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_item import Include, Inherit, Item, Variable
from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class VarDependsAppend(Rule):
    def __init__(self) -> None:
        super().__init__(id='oelint.vars.dependsappend',
                         severity='error',
                         message='DEPENDS should only be appended, not overwritten as an include or inherit')

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        items: List[Variable] = stash.GetItemsFor(filename=_file, classifier=Variable.CLASSIFIER,
                                                  attribute=Variable.ATTR_VAR, attributeValue='DEPENDS')
        incinh: List[Item] = stash.GetItemsFor(filename=_file, classifier=[
                                               Inherit.CLASSIFIER, Include.CLASSIFIER], nolink=True)

        earliest = 99999999
        if incinh:
            earliest = min(x.Line for x in incinh)

        for i in items:
            if not i.AppendOperation() and i.Line > earliest:
                res += self.finding(i.Origin, i.InFileLine)
        return res



from typing import List, Tuple

from oelint_parser.cls_stash import Stash

from oelint_adv.cls_rule import Rule


class NewLineConsecutive(Rule):
    def __init__(self):
        super().__init__(id='oelint.newline.consecutive',
                         severity='warning',
                         message='Consecutive blank lines should be avoided')

    def __getMatches(self, _file: str, stash: Stash) -> dict:
        res = {}
        items = stash.GetItemsFor(filename=_file)
        for _uniqname in {x.Origin for x in items}:
            res[_uniqname] = sorted(stash.GetItemsFor(
                filename=_uniqname, nolink=True), key=lambda x: x.InFileLine)
        return res

    def check(self, _file: str, stash: Stash) -> List[Tuple[str, int, str]]:
        res = []
        for _, v in self.__getMatches(_file, stash).items():
            for index, value in enumerate(v):
                if index == 0:
                    continue
                if value.Raw == '\n' and v[index - 1].Raw == '\n':
                    res += self.finding(value.Origin, value.InFileLine)
        return res

    def fix(self, _file: str, stash: Stash) -> List[str]:
        res = set()
        for f, v in self.__getMatches(_file, stash).items():
            for index, value in enumerate(v):
                if index == 0:
                    continue
                if value.RealRaw == '\n' and v[index - 1].RealRaw == '\n':
                    stash.Remove(value)
                    res.add(f)
        return list(res)
